
structure PolyMinML =
struct

    open Types

    datatype primop =
        PLUS
      | TIMES
      | MINUS
      | NEGATE

    fun primoptostring PLUS = "+"
      | primoptostring TIMES = "*"
      | primoptostring MINUS = "-"
      | primoptostring NEGATE = "~"

    local 
        val c = ref 0
        fun ++ i = (i := (!i + 1); !i)
    in
        fun newvar () = "v" ^ Int.toString (++ c)
    end

    infix ++

    datatype cexp =
        I of iexp
      | CFun of var * var * cexp
      | CTypefun of var * cexp
      | ....  (* fill this in *)

    and iexp =
        Unit
      | Var of string
      | Int of int
      | Bool of bool
      | Primop of primop * cexp list
      | Fun of var * var * typ * typ * cexp
      | Typefun of var * iexp
      | ....... (* fill this in *)

    fun TypeofPrimop PLUS = ([INT, INT], INT)
      | TypeofPrimop TIMES = ([INT, INT], INT)
      | TypeofPrimop MINUS = ([INT, INT], INT)
      | TypeofPrimop NEGATE = ([INT], INT)

    exception PrimopError

    fun EvalPrimop (PLUS, [Int a, Int b])  = Int (a + b)
      | EvalPrimop (MINUS, [Int a, Int b]) = Int (a - b)
      | EvalPrimop (TIMES, [Int a, Int b]) = Int (a * b)
      | EvalPrimop (NEGATE, [Int a])       = Int (~ a)
      | EvalPrimop _ = raise PrimopError

    (* substitute types within a cexp *)
    fun ctsubst S (CFun(f, x, c)) = CFun(f, x, ctsubst S c)
      | ctsubst S (CTypefun (v, c)) = CTypefun (v, ctsubst (S ++ (v, VAR v)) c)
      | ctsubst S (I i) = I (itsubst S i)
      | ... (* fill this in *)

    (* substitute types within an iexp *)
    and itsubst S (Primop (p, cl)) = Primop (p, map (ctsubst S) cl)
      | itsubst S (Fun (f, x, td, tr, c)) = Fun (f, x, (tsubst S td), (tsubst S tr), ctsubst S c)
      | itsubst S (Inst (i, t)) = Inst (itsubst S i, (tsubst S t))
      | itsubst S (Typefun (v, i)) = Typefun (v, itsubst (S ++ (v, VAR v)) i)
      | .... (* fill this in *)
      | itsubst S other = other


end